@@ -1,4 +1,4 @@
-from typing_extensions import AsyncIterable, TypeVar, Generic, Callable
+from typing_extensions import AsyncIterable, TypeVar, Generic, Callable, overload
 from pydantic import RootModel
 from haskellian import Either, Left, Right, promise as P, asyn_iter as AI
 from sqlalchemy import Engine
@@ -9,11 +9,12 @@ from sqltypes import ValidatedJSON
 from kv import KV, DBError, InexistentItem, InvalidData
 
 T = TypeVar('T')
+U = TypeVar('U')
 
 class SQLKV(KV[T], Generic[T]):
   """`KV` implementation over sqlalchemy"""
 
-  def __init__(self, Type: type[T], engine: Callable[[], Engine], table: str = 'kv'):
+  def __init__(self, Type: type[T], engine: Engine, *, table: str):
     self.engine = engine
 
     class Base(DeclarativeBase):
@@ -45,13 +46,30 @@ class SQLKV(KV[T], Generic[T]):
         value: Mapped[RootModel[Type]] = mapped_column(type_=ValidatedJSON(Root))
 
     self.Table = Table
-    Base.metadata.create_all(engine())
+    Base.metadata.create_all(engine)
+
+  def __repr__(self):
+    return f'SQLKV(engine={self.engine!r}, table={self.Table.__tablename__!r}, type={self.Table.value.type!r})'
+
+  @overload
+  @classmethod
+  def new(cls, conn_str: str, *, table: str) -> 'SQLKV[bytes]':
+    ...
+  @overload
+  @classmethod
+  def new(cls, conn_str: str, type: type[U] | None = None, *, table: str) -> 'SQLKV[U]':
+    ...
+  @classmethod
+  def new(cls, conn_str: str, type: type[U] | None = None, *, table: str): # type: ignore
+    from sqlalchemy import create_engine
+    engine = create_engine(conn_str)
+    return cls(type or bytes, engine, table=table)
 
   @P.lift
   async def delete(self, key: str) -> Either[DBError | InexistentItem, None]:
     from sqlmodel import Session, select
     try:
-      with Session(self.engine()) as session:
+      with Session(self.engine) as session:
         stmt = select(self.Table).where(self.Table.key == key)
         row = session.exec(stmt).first()
         if row is None:
@@ -66,7 +84,7 @@ class SQLKV(KV[T], Generic[T]):
   async def read(self, key: str) -> Either[DBError | InvalidData | InexistentItem, T]:
     from sqlmodel import Session, select
     try:
-      with Session(self.engine()) as session:
+      with Session(self.engine) as session:
         stmt = select(self.Table).where(self.Table.key == key)
         row = session.exec(stmt).first()
         if row is None:
@@ -79,7 +97,7 @@ class SQLKV(KV[T], Generic[T]):
   async def insert(self, key: str, value: T) -> Either[DBError, None]:
     from sqlmodel import Session, select
     try:
-      with Session(self.engine()) as session:
+      with Session(self.engine) as session:
         stmt = select(self.Table).where(self.Table.key == key)
         row = session.exec(stmt).first()
         if row is not None:
@@ -94,9 +112,9 @@ class SQLKV(KV[T], Generic[T]):
   async def keys(self) -> AsyncIterable[Either[DBError, str]]:
     from sqlmodel import Session, select
     try:
-      with Session(self.engine()) as session:
+      with Session(self.engine) as session:
         stmt = select(self.Table.key)
-        for key in session.exec(stmt):
+        for key in session.exec(stmt).all():
           yield Right(key)
     except DatabaseError as e:
       yield Left(DBError(e))
@@ -105,7 +123,7 @@ class SQLKV(KV[T], Generic[T]):
   async def items(self, batch_size: int | None = None) -> AsyncIterable[Either[DBError | InvalidData, tuple[str, T]]]:
     from sqlmodel import Session, select
     try:
-      with Session(self.engine()) as session:
+      with Session(self.engine) as session:
         result = session.exec(select(self.Table))
         while (batch := result.fetchmany(batch_size)) != []:
           for row in batch:
@@ -118,7 +136,7 @@ class SQLKV(KV[T], Generic[T]):
   async def clear(self):
     from sqlmodel import Session, delete
     try:
-      with Session(self.engine()) as session:
+      with Session(self.engine) as session:
         session.execute(delete(self.Table))
         session.commit()
         return Right(None)
